//-------------------------------------------------------BEGINHEADER--
// This file is part of LoopSim (arXiv:1006.2144).
//
// Copyright 2009-2011 Mathieu Rubin, Gavin P. Salam and Sebastian Sapeta
//
// This software is provided as a snapshot of the ongoing LoopSim
// research project. As such, you may not redistribute it without the
// authors' permission.
//
// If you use LoopSim as part of your scientific work, you should
// discuss and agree with the LoopSim authors how best to acknowledge
// LoopSim in your work (e.g. whether through a reference, or through
// joint authorship with the LoopSim authors). 
//
// To help guide LoopSim's proper use in its current early stage of
// development, a condition of use of LoopSim is that any results that
// you obtain with it must be shown to the LoopSim authors before
// they are made public.
//
// $Id: ExtractLHEvent.cc 1918 2013-06-29 14:07:12Z sapeta $
//-------------------------------------------------------ENDHEADER----

#include "ExtractLHEvent.hh"


//-----------------------------------------------------------------------------
// constructor
ExtractLHEvent::ExtractLHEvent() : _generator(NONE), 
                                   _reading_event_block(false), 
                                   _current_event(0), _ientry(-1),
				   _end_of_file(false), _n_bosons(0),
				   _n_leptons(0) {
}


//-----------------------------------------------------------------------------
// read each line and take action depending on its content
void ExtractLHEvent::readline(string line, LHEFormat lheformat) {

  // new event encountered, set flags so as to read the following lines
  // to the new _current_event_block
  if (line.find("<event>")!=string::npos) {
    _reading_event_block = true;
    _current_event_block.str("");
    _current_event.reset(0);
  // end of an event; now the event is complete and we call function 
  // which analyses it
  } else if (line.find("</event>")!=string::npos) {
    _reading_event_block = false;
    switch (lheformat) {
      case vbfnlo:
      _analyse_event_vbfnlo();
      break;
      case with_extra_info:
      _analyse_event();
      break;
    }
  // middle of an event; read the next line and save it in the
  // _current_event_blck
  } else if (_reading_event_block) {
    _current_event_block << line << endl;
  } else {
    _current_event.reset(0);
  }

  // set the appropriate flag when end of the LHEF input is reached
  if (line.find("</LesHouchesEvents>")!=string::npos) {
   _end_of_file = true; 
  }

  // get info about generator and store it
  if (line.find("MCFM")!=string::npos) {
    _generator = MCFM;
  } else if (line.find("VBFNLO")!=string::npos) {
    _generator = VBFNLO;
  }

}

//-----------------------------------------------------------------------------
// analyze a comlete event, i.e. take the information stored in LHE format and
// transfer it into the LoopSim Event class
//
// - possible bosonic final states: Z, W+, W-, W+Z, W-Z
// - the above can dacay to either electron or muon flavour
//
void ExtractLHEvent::_analyse_event() {
   ++_ientry;
   _current_event.reset(new Event());
   // LHE  block consists of a number of variables which we read now;
   // for more explanation take a look at hep-ph/0609017 and hep-ph/0109068
   string idprup, comchar, part;
   int nup, idup, istup, nd, it;
   double facscale, aemmz, as, mothup1, mothup2,  icolup1, icolup2,
          m, vtimup, spinup;
   double xwgtup, px, py, pz, E;
   // read event information
   _current_event_block >> nup >>  idprup >> xwgtup >> facscale >>  aemmz >> as;
   // save the event weight
   _current_event->weight  = xwgtup;

   // read information for each particle and store it in LS Event class
   vector<PseudoJet> eminus, eplus, muminus, muplus, nue, antinue, numu, 
                     antinumu;

   // keep track of whether we might have missed particles
   _n_unused = 0;

   for (int i=0;i<nup;i++) {
     _current_event_block >> idup >>  istup >>  mothup1 >>  mothup2 >>  icolup1
                          >> icolup2 >> px >> py >> pz >> E >> m >> vtimup 
			  >> spinup;
     // skip beam particles and intermediate particles
     // we are interested only in the final state particles after decay
     // to be removed; just checking if this condition can still be satisfied
     // which it should not
     if (istup != 1 ) continue;
     //if (istup < 0 ) { exit(-1);}

     // so far we've not used the particle
     _n_unused++;

     // make sure the particle has a user index (so we can track it in decays)
     PseudoJet particle(px,py,pz,E);
     particle.set_user_index(idup);

     if (idup == 11) {
       eminus.push_back(PseudoJet(px,py,pz,E));
       eminus[0].set_user_index(11);
     } else if (idup == -11) {
       eplus.push_back(PseudoJet(px,py,pz,E));
       eplus[0].set_user_index(-11);
     } else if (idup == 13) {
       muminus.push_back(PseudoJet(px,py,pz,E));
       muminus[0].set_user_index(13);
     } else if (idup == -13) {
       muplus.push_back(PseudoJet(px,py,pz,E));
       muplus[0].set_user_index(-13);
     } else if (idup == 12) {
       nue.push_back(PseudoJet(px,py,pz,E));
       nue[0].set_user_index(12);
     } else if (idup == -12) {
       antinue.push_back(PseudoJet(px,py,pz,E));
       antinue[0].set_user_index(-12);
     } else if (idup == 14) {
       numu.push_back(PseudoJet(px,py,pz,E));
       numu[0].set_user_index(14);
     } else if (idup == -14) {
       antinumu.push_back(PseudoJet(px,py,pz,E));
       antinumu[0].set_user_index(-14);
     }

     assert (eminus.size() <= 1 || eplus.size() <= 1 ||
             muminus.size() <= 1 || muplus.size() <= 1 ||
             nue.size() <= 1 || antinue.size() <= 1 ||
             numu.size() <= 1 || antinumu.size() <= 1 );

     // check if found Z or W+/W- boson
     _check_boson(23,eminus,eplus);
     _check_boson(23,muplus,muminus);
     _check_boson(24,eplus,nue);
     _check_boson(-24,eminus,antinue);
     _check_boson(24,muplus,numu);
     _check_boson(-24,muminus,antinumu);

     // a parton
     if (idup == 81 || idup == 21 || (abs(idup) >= 1 && abs(idup) <= 6)) {
       _current_event->particles.push_back(LSParticle(px,py,pz,E,81));
       _n_unused--;
     } else if (idup == 22) {
       // a photon 
       _current_event->particles.push_back(LSParticle(px,py,pz,E,22));
       _n_unused--;
     }
   }


   if (_n_leptons/_n_bosons != 2)
     throw_error_message(
     "the number of final state leptons does not match the number of bosons",
     _current_event_block.str());


   // make sure we don't have any unused particles left over
   assert(_n_unused == 0);

   // read optional information about the event 
   // (optional for LHEF but for us that is necessary)
   // part: lord, real, virt
   // nd  : dipole number, nd=0 is the real no-dipole term
   // it  : iteration number
   _current_event_block >> comchar >> part >> nd >> it;
   _current_event_extra_info.set_part(part);
   _current_event_extra_info.set_nd(nd);
   _it = it;
   if(comchar != "#") {
     throw_error_message("optional line does not start with #",
                         _current_event_block.str());
   }
   // if the event is a dipole remove spurious particle which is assumed 
   // to have 4-momentum = (0,0,0,0) which seems to be the case for vbfnlo
   if (_current_event_extra_info.is_a_dipole()) {
     double some_eps = 1e-7;
     unsigned i;
     for (i=0; i<_current_event->particles.size();i++) {
       if (abs(_current_event->particles[i].E() )< some_eps &&  
           abs(_current_event->particles[i].px())< some_eps &&  
           abs(_current_event->particles[i].py())< some_eps &&  
           abs(_current_event->particles[i].pz())< some_eps)
	   break;
     }
     _current_event->particles.erase(_current_event->particles.begin()+i);
   }

   //cout << _current_event_block.str() << endl;
   //print_event(*_current_event);
}

//-----------------------------------------------------------------------------
void ExtractLHEvent::_check_boson(int flav, vector<PseudoJet>& lepton1, 
                                            vector<PseudoJet>& lepton2) {

  if (lepton1.size() == 1 && lepton2.size() == 1) {
    PseudoJet boson = lepton1[0]+lepton2[0];
    vector<PseudoJet> leptons;
    leptons.push_back(lepton1[0]);
    leptons.push_back(lepton2[0]);
    _current_event->particles.push_back(LSParticle(boson,flav,leptons));
    _n_unused -= 2;
    lepton1.resize(0);
    lepton2.resize(0);
    _n_bosons +=1;
    _n_leptons+=2;
  }
}

//-----------------------------------------------------------------------------
// true if we have just finished one iteration and started another one
// this is when one should in principle write intemediate results to get them
// with correct normalization
// 
bool ExtractLHEvent::periodic_output() {

  static int it_old = 1;
  if (_it> it_old) {
    it_old++;
    return true;
  } else {
    return false;
  }
}

//-----------------------------------------------------------------------------
// simple function to throwing an error message
void ExtractLHEvent::throw_error_message(string message, string explanation) {

  cerr << "ExtractLHEvent::ERROR: " << message << endl;
  cerr << explanation << endl;
  exit(-1);
}

//-----------------------------------------------------------------------------
//
void ExtractLHEvent::_analyse_event_vbfnlo() {
   // this function will soon be removed
   cerr << "--------------------------------------------------------" << endl;
   cerr << "The _analyse_event_vbfnlo() function is obsolete now" <<endl;
   cerr << "and is here just for tests. If you really want to use it," << endl;
   cerr << "uncomment the exits statement in the code." << endl;
   cerr << "--------------------------------------------------------" << endl;
   exit(1);
   ++_ientry;
   _current_event.reset(new Event());
   // LHE  block consists of a number of variables which we read now;
   // for more explanation take a look at hep-ph/0609017 and hep-ph/0109068
   string idprup, comchar, part;
   int nup, idup, istup, nd, it;
   double facscale, aemmz, as, mothup1, mothup2,  icolup1, icolup2,
          m, vtimup, spinup;
   double xwgtup, px, py, pz, E;
   // read event information
   _current_event_block >> nup >>  idprup >> xwgtup >> facscale >>  aemmz >> as;
   // save the event weight
   _current_event->weight  = xwgtup;

   // read information for each particle and store it in LS Event class
   vector<PseudoJet> eminus, eplus, muminus, muplus, nue, antinue, numu, 
                     antinumu;

   for (int i=0;i<nup;i++) {
     _current_event_block >> idup >>  istup >>  mothup1 >>  mothup2 >>  icolup1
                          >> icolup2 >> px >> py >> pz >> E >> m >> vtimup 
			  >> spinup;
     // skip beam particles
     // to be removed; just checking if this condition can still be satisfied
     // which it should not
     if (istup < 0 ) continue;

     if (idup == 11) {
       eminus.push_back(PseudoJet(px,py,pz,E));
       eminus[0].set_user_index(11);
     } else if (idup == -11) {
       eplus.push_back(PseudoJet(px,py,pz,E));
       eplus[0].set_user_index(-11);
     } else if (idup == 13) {
       muminus.push_back(PseudoJet(px,py,pz,E));
       muminus[0].set_user_index(13);
     } else if (idup == -13) {
       muplus.push_back(PseudoJet(px,py,pz,E));
       muplus[0].set_user_index(-13);
     } else if (idup == 12) {
       nue.push_back(PseudoJet(px,py,pz,E));
       nue[0].set_user_index(12);
     } else if (idup == -12) {
       antinue.push_back(PseudoJet(px,py,pz,E));
       antinue[0].set_user_index(-12);
     } else if (idup == 14) {
       numu.push_back(PseudoJet(px,py,pz,E));
       numu[0].set_user_index(14);
     } else if (idup == -14) {
       antinumu.push_back(PseudoJet(px,py,pz,E));
       antinumu[0].set_user_index(-14);
     }

     assert (eminus.size() <= 1 || eplus.size() <= 1 ||
             muminus.size() <= 1 || muplus.size() <= 1 ||
             nue.size() <= 1 || antinue.size() <= 1 ||
             numu.size() <= 1 || antinumu.size() <= 1 );

     // check if found Z or W+/W- boson
     _check_boson(23,eminus,eplus);
     _check_boson(23,muplus,muminus);
     _check_boson(24,eplus,nue);
     _check_boson(-24,eminus,antinue);
     _check_boson(24,muplus,numu);
     _check_boson(-24,muminus,antinumu);

     // a parton
     if (idup == 81 || idup == 21 || (abs(idup) >= 1 && abs(idup) <= 6))
       _current_event->particles.push_back(LSParticle(px,py,pz,E,81));
   }

   if (_n_leptons/_n_bosons != 2)
     throw_error_message("the number of final state leptons does not match the number of bosons", _current_event_block.str());


   // read optional information about the event 
   // (optional for LHEF but for us that is necessary)
   // part: lord, real, virt
   // nd  : dipole number, nd=0 is the real no-dipole term
   // it  : iteration number
   _current_event_block >> comchar >> part >> nd >> it;
   _current_event_extra_info.set_part(part);
   _current_event_extra_info.set_nd(nd);
   _it = it;
   if(comchar != "#") {
     throw_error_message("optional line does not start with #",
                         _current_event_block.str());
   }
   // if the event is a dipole remove spurious particle which is assumed 
   // to have 4-momentum = (0,0,0,0) which seems to be the case for vbfnlo
   if (_current_event_extra_info.is_a_dipole()) {
     double some_eps = 1e-7;
     unsigned i;
     for (i=0; i<_current_event->particles.size();i++) {
       if (abs(_current_event->particles[i].E() )< some_eps &&  
           abs(_current_event->particles[i].px())< some_eps &&  
           abs(_current_event->particles[i].py())< some_eps &&  
           abs(_current_event->particles[i].pz())< some_eps)
	   break;
     }
     _current_event->particles.erase(_current_event->particles.begin()+i);
   }
}


//-----------------------------------------------------------------------
string ExtractLHEvent::generator() const {
  if (_ientry < 0) {
    cerr << "-- ERROR  -----------------------------------------------" << endl;
    cerr << "It is too early to use this function as not all necessary"  << endl; 
    cerr << "data has been read from the LHE file. Read at least one "   << endl;
    cerr << "event before checking generator name!"                      << endl;
    exit(1);
  }
  if (_generator == MCFM) {
    return "MCFM";
  } else if (_generator == VBFNLO) {
    return "VBFNLO";
  } else if (_generator == NONE) {
    return "NONE";
  } else {
    return "Not even NONE! What is your generator then?";
  }
}
