//-------------------------------------------------------BEGINHEADER--
// This file is part of LoopSim (arXiv:1006.2144).
//
// Copyright 2009-2011 Mathieu Rubin, Gavin P. Salam and Sebastian Sapeta
//
// This software is provided as a snapshot of the ongoing LoopSim
// research project. As such, you may not redistribute it without the
// authors' permission.
//
// If you use LoopSim as part of your scientific work, you should
// discuss and agree with the LoopSim authors how best to acknowledge
// LoopSim in your work (e.g. whether through a reference, or through
// joint authorship with the LoopSim authors). 
//
// To help guide LoopSim's proper use in its current early stage of
// development, a condition of use of LoopSim is that any results that
// you obtain with it must be shown to the LoopSim authors before
// they are made public.
//
// $Id: InputDat.hh 1919 2013-06-29 15:25:14Z sapeta $
//-------------------------------------------------------ENDHEADER----

// issues:
//
//   * there is some problem with reading options having underscore in them
//
#ifndef __INPUTDAT_HH__
#define __INPUTDAT_HH__

#include <cstdlib>
#include <string>
#include <map>
#include <vector>
#include <fstream>
#include <sstream>

using namespace std;

/// Class for reading options from card file.
///
/// The input file is assumed in the form
///
///   [option_name] value
///
/// Any number of lines is allowed. The options and their values are stored in
/// the map. To retrieve them one used a template function 'value' with an
/// appropriate type
///
///   inputdat->value<int>("ptmin0")
///
class InputDat {

  public:
  InputDat(string filename) : _filename(filename) {
    _read_options();
  }
 
  InputDat(const InputDat & other) {
    _filename = other._filename;
    _options = other._options;
  }


  /// return value corresponding to a given option name
  template<class T> T value (const string& name) {

    /// set iterator to the key='name'
    /// if there is no entry corresponding entry, throw an error
    map<string,string>::iterator it=_options.find("["+name+"]");
    if( it == _options.end())  {
      cerr << "InputDat::ERROR: There is no option corresponding to the name " 
           << name << endl;
      exit(-1);
    } 
    T result;
    istringstream istream((*it).second);
    //istringstream istream(_options["["+name+"]"]);
    istream >> result;
    return result;
  }

  /// return list of all options and values
  string description() {
    stringstream sstr;
    sstr <<  "# Input file " << _filename << endl;
    for (map<string,string>::iterator it=_options.begin();
         it != _options.end();it++) 
      sstr <<  "# " << (*it).first << " " << (*it).second << endl;
    return sstr.str();
  }

  /// return the name of the card file
  string filename () const { return _filename;}
  private:
  string _filename;
  map<string, string> _options;
  
  /// read options from file and save them in this class
  void _read_options() {

    ifstream indata;
    indata.open(_filename.c_str());
    if(!indata) {
       cerr << "InputDat::ERROR: file " << _filename.c_str() 
            << " could not be opened" << endl;
       exit(1);
    }
    while(true) {
      string name, value;
      indata >> name >> value;
      if (!indata.good()) break;
      _options[name]=value;
    }
    indata.close();
  }
};

#endif // __INPUTDAT_HH__
